/**
  *  \file client/tiles/fleetoverview.cc
  *  \brief WFleetOverviewTile tile
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/fleetoverview.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/mission.h"
#include "game/ship.h"
#include "game/univ.h"
#include "gfx/gfxcontext.h"
#include "ui/uiglobal.h"
#include "util/nls.h"

WFleetOverviewTile::WFleetOverviewTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(font_sizes[FONT_NORMAL_BOLD].scaledBy(30, 10), selection)
{ }

WFleetOverviewTile::~WFleetOverviewTile()
{ }

void
WFleetOverviewTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    drawBackground(ctx, getInnerSize(0, 0));

    GShip* psh = dynamic_cast<GShip*>(getCurrentObject());
    if (!psh || psh->getFleetNumber() != psh->getId()) {
        return;
    }

    // Draw it
    GUniverse& univ = *getCurrentUniverse();
    GShip& sh = *psh;

    int y = 0, x = 0;
    int w = getInnerSize(0, 0).w;
    string_t text;

    // Line 1: Name
    text = format(_("Leader of fleet #%d"), sh.getId());
    if (!sh.getFleetName().empty()) {
        text += ": ";
        text += sh.getFleetName();
    }
    outTextF(ctx, x, y, w, text);
    y += font_heights[FONT_NORMAL];

    // Line 2: Location
    outTextF(ctx, x, y, w, _("Location: ") + univ.getLocationName(sh.getPos(), locs_WW | locs_Orbit | locs_Verbose));
    y += font_heights[FONT_NORMAL];

    // Line 3: Waypoint
    const GMission*const msn = GMissionList::getInstance().getMissionByShip(sh);
    if (msn != 0 && (msn->getFlags() & GMission::mf_Waypoint) != 0) {
        outTextF(ctx, x, y, w, _("Mission: ") + msn->getLabel(sh));
    } else {
        outTextF(ctx, x, y, w, _("Waypoint: ") + univ.getLocationName(sh.getWaypoint(), locs_WW | locs_Verbose));
    }
    y += 3*font_heights[FONT_NORMAL]/2;

    // Line 4ff: Ship counts
    int fre = 0, cap = 0, tow = 0;
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& si = univ.getShip(i);
        if (si.getFleetNumber() == sh.getId()) {
            if (si.getNumBeams() > 0 || si.getNumTorpLaunchers() > 0 || si.getNumBays() > 0) {
                ++cap;
            } else {
                ++fre;
            }
            if (si.getMission() == GMission::msn_Tow
                && univ.isValidShipId(si.getTowId())
                && si.getTowId() != si.getId()
                && univ.getShip(si.getTowId()).getFleetNumber() == sh.getId())
            {
                ++tow;
            }
        }
    }
    if (cap) {
        outTextF(ctx, x, y, w, format(_("%d capital ship%!1{s%}"), cap));
        y += font_heights[FONT_NORMAL];
    }
    if (fre) {
        outTextF(ctx, x, y, w, format(_("%d freighter%!1{s%}"), fre));
        y += font_heights[FONT_NORMAL];
    }
    if (tow) {
        outTextF(ctx, x, y, w, format(_("%d fleet member%1{ is%|s are%} being towed."), tow));
        y += font_heights[FONT_NORMAL];
    }
}
