/**
  *  \file game/planet-hist.cc
  *  \brief Planets: History Information
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/planet.h"
#include "game/parser.h"
#include "game/planet-int.h"
#include "util/recode.h"

static void
updateField16(int16_t& field_time, int16_t time,
              int16_t& field_value, int32_t value)
{
    if (field_time <= time || !mp16_t(field_value).isKnown()) {
        field_value = value;
        if (field_time < time)
            field_time = time;
    }
}

static void
updateField32(int16_t& field_time, int16_t time,
              int32_t& field_value, int32_t value)
{
    if (field_time <= time || !mp32_t(field_value).isKnown()) {
        field_value = value;
        if (field_time < time)
            field_time = time;
    }
}

static void
updateAdd32(int16_t& field_time, int16_t time,
            int32_t& field_value, int32_t added)
{
    /* We want this to be idempotent in some way, so we cannot add on every
       iteration we go through this. Since these reports come from meteorites,
       we can assume only one such report per planet. */
    if (!mp32_t(field_value).isKnown()) {
        field_value = added;
        field_time = time;
    } else if (field_time < time) {
        field_value += added;
        field_time = time;
    }
}

/** Check acceptance of message information.
    We must filter information so we don't by accident overwrite good, reliable
    information with information from a scan. */
static bool
acceptMessageInformation(const GPlanet& pl, const GMessageValueBase& info)
{
    // White-list of items that are always accepted
    if (const GMessageIntValue* iv = dynamic_cast<const GMessageIntValue*>(&info)) {
        if (iv->getIndex() == mi_BaseQueuePos || iv->getIndex() == mi_BaseQueuePriority) {
            return true;
        }
    }

    // Everything else is only accepted when we're in history
    return pl.getPlanetSource().empty();
}

/** Add message ifnormation. Process information received from messages or util.dat. */
void
GPlanet::addMessageInformation(const GMessageInformation& info, int msgNr)
{
    /* now process the information. For timestamp handling, we assume
       that information comes in in full form, and in sequential order.
       During normal operation, the timestamp checks will always succeed,
       as pl.time[] starts with a previous turn number and, since
       information comes in sequential order, msg_turn is either the
       current turn number or the one before. */
    const int16_t msg_turn = info.getTurnNumber();
    PlanetInfo& pl = createPlanetInfo();
    for (GMessageInformation::iterator i = info.begin(); i != info.end(); ++i) {
        if (!acceptMessageInformation(*this, **i)) {
            // ignore
        } else if (GMessageStringValue* sv = dynamic_cast<GMessageStringValue*>(*i)) {
            switch (sv->getIndex()) {
             case ms_FCode:
                // FCode always comes with an industry report, so associate it
                // with colonists.
                if (pl.time[ts_Colonists] <= msg_turn || pl.data.fcode[0] == UNKNOWNC) {
                    storeBasicString(pl.data.fcode, convertUtf8ToGame(sv->getValue()));
                    if (pl.time[ts_Colonists] <= msg_turn)
                        pl.time[ts_Colonists] = msg_turn;
                }
                break;
             default:
                break;
            }
        } else if (GMessageIntValue* iv = dynamic_cast<GMessageIntValue*>(*i)) {
            switch (iv->getIndex()) {
             case mi_X:
             case mi_Y:
                // Is it useful to support these for planets?
                break;
             case mi_Owner:
                updateField16(pl.time[ts_Colonists], msg_turn, pl.data.owner, iv->getValue());
                break;
             case mi_PlanetTotalN:
                // Total is reported by Dark Sense. Treat it as Ground.
                updateField32(pl.time[ts_Minerals], msg_turn, pl.data.ore_ground[0], iv->getValue());
                break;
             case mi_PlanetTotalT:
                updateField32(pl.time[ts_Minerals], msg_turn, pl.data.ore_ground[1], iv->getValue());
                break;
             case mi_PlanetTotalD:
                updateField32(pl.time[ts_Minerals], msg_turn, pl.data.ore_ground[2], iv->getValue());
                break;
             case mi_PlanetTotalM:
                updateField32(pl.time[ts_Minerals], msg_turn, pl.data.ore_ground[3], iv->getValue());
                break;
             case mi_PlanetAddedN:
                updateAdd32(pl.time[ts_Minerals], msg_turn, pl.data.ore_ground[0], iv->getValue());
                break;
             case mi_PlanetAddedT:
                updateAdd32(pl.time[ts_Minerals], msg_turn, pl.data.ore_ground[1], iv->getValue());
                break;
             case mi_PlanetAddedD:
                updateAdd32(pl.time[ts_Minerals], msg_turn, pl.data.ore_ground[2], iv->getValue());
                break;
             case mi_PlanetAddedM:
                updateAdd32(pl.time[ts_Minerals], msg_turn, pl.data.ore_ground[3], iv->getValue());
                break;
             case mi_PlanetMinedN:
                updateField32(pl.time[ts_Minerals], msg_turn, pl.data.ore_mined[0], iv->getValue());
                break;
             case mi_PlanetMinedT:
                updateField32(pl.time[ts_Minerals], msg_turn, pl.data.ore_mined[1], iv->getValue());
                break;
             case mi_PlanetMinedD:
                updateField32(pl.time[ts_Minerals], msg_turn, pl.data.ore_mined[2], iv->getValue());
                break;
             case mi_PlanetMinedM:
                updateField32(pl.time[ts_Minerals], msg_turn, pl.data.ore_mined[3], iv->getValue());
                break;
             case mi_PlanetCash:
                updateField32(pl.time[ts_Cash], msg_turn, pl.data.money, iv->getValue());
                break;
             case mi_PlanetSupplies:
                updateField32(pl.time[ts_Cash], msg_turn, pl.data.supplies, iv->getValue());
                break;
             case mi_PlanetHasBase:
                updateField16(pl.time[ts_Colonists], msg_turn, pl.data.build_base, iv->getValue());
                break;
             case mi_PlanetMines:
                updateField16(pl.time[ts_Colonists], msg_turn, pl.data.mines, iv->getValue());
                break;
             case mi_PlanetFactories:
                updateField16(pl.time[ts_Colonists], msg_turn, pl.data.factories, iv->getValue());
                break;
             case mi_PlanetDefense:
                updateField16(pl.time[ts_Colonists], msg_turn, pl.data.defense, iv->getValue());
                break;
             case mi_PlanetTemperature:
                // Temperature has no timestamp field (because it changes only very seldom),
                // and always comes with a Colonists or Natives report (explore/bioscan).
                if (!mp16_t(pl.data.temp_code).isKnown()
                    || msg_turn >= pl.time[ts_Colonists]
                    || msg_turn >= pl.time[ts_Natives])
                {
                    pl.data.temp_code = 100 - iv->getValue();
                }
                break;
             case mi_PlanetColonists:
                updateField32(pl.time[ts_Colonists], msg_turn, pl.data.colonists, iv->getValue());
                break;
             case mi_PlanetActivity:
                if (msg_turn >= pl.time[ts_Colonists] || !pl.industry_level.isKnown()) {
                    setIndustryLevel(mp16_t(iv->getValue()));
                    if (msg_turn >= pl.time[ts_Colonists])
                        pl.time[ts_Colonists] = msg_turn;
                }
                break;
             case mi_PlanetNativeRace:
                updateField16(pl.time[ts_Natives], msg_turn, pl.data.native_race, iv->getValue());
                if (iv->getValue() == 0) {
                    // Report of native race 0 means natives are gone
                    updateField16(pl.time[ts_Natives], msg_turn, pl.data.native_gov, 0);
                    updateField32(pl.time[ts_Natives], msg_turn, pl.data.natives,    0);
                }
                break;
             case mi_PlanetNativeGov:
                updateField16(pl.time[ts_Natives], msg_turn, pl.data.native_gov, iv->getValue());
                break;
             case mi_PlanetNatives:
                updateField32(pl.time[ts_Natives], msg_turn, pl.data.natives, iv->getValue());
                if (iv->getValue() == 0) {
                    // Report of population 0 means natives are gone
                    updateField16(pl.time[ts_Natives], msg_turn, pl.data.native_race, 0);
                    updateField16(pl.time[ts_Natives], msg_turn, pl.data.native_gov,  0);
                }
                break;
             case mi_PlanetHasNatives:
                if (msg_turn >= pl.time[ts_Natives]) {
                    setKnownToHaveNatives(iv->getValue());
                    if (msg_turn >= pl.time[ts_Natives])
                        pl.time[ts_Natives] = msg_turn;
                }
                break;
             case mi_BaseQueuePos:
                setBaseQueuePosition(mp16_t(iv->getValue()));
                break;
             case mi_BaseQueuePriority:
                setBaseQueuePriority(mp32_t(iv->getValue()));
                break;
             default:
                break;
            }
        }
    }

    /* Associate message */
    if (msgNr >= 0) {
        if (associated_messages.empty() || associated_messages.back() != msgNr) {
            associated_messages.push_back(msgNr);
        }
    }

    /* mark planet dirty, just in case */
    markDirty();
}
