/**
  *  \file int/if/baseif.cc
  *  \brief Interpreter Game Interface: Starbase Properties
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/baseif.h"
#include "cpluslib/format.h"
#include "game/actions/baseammo.h"
#include "game/actions/basefix.h"
#include "game/actions/baseparts.h"
#include "game/actions/baseship.h"
#include "game/actions/basetech.h"
#include "game/config.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/tables.h"
#include "game/undo.h"
#include "int/args.h"
#include "int/compound.h"
#include "int/exec.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/value.h"
#include "game/planetform.h"
#include "game/planettransfer.h"
#include "game/shiptransfer.h"
#include "game/actions/preconditions.h"

namespace {
    /** Starbase array property. This handles a starbase's array properties, Storage.Xxx(). */
    class IntBaseArrayProperty : public IntIndexableValue {
        const GPlanet& p;
        const IntBaseProperty ibp;

        IntValue* performArrayReference(mp16_t (GPlanet::*func)(int) const, int32_t limit, int32_t arg, bool hull);
     public:
        IntBaseArrayProperty(const GPlanet& p, IntBaseProperty ibp);

        // IntIndexableValue:
        IntValue* get(IntArgBlock& args);
        void set(IntArgBlock& args, IntValue* value);
        int32_t getDimension(int32_t dim);
        IntContext* makeFirstContext();

        // IntValue:
        IntBaseArrayProperty* clone();
        string_t toString(bool readable);
        void store(IntTagNode& sv, Stream& aux);
    };
}

/** Constructor.
    \param p Starbase
    \param ibp Property to implement */
IntBaseArrayProperty::IntBaseArrayProperty(const GPlanet& p, IntBaseProperty ibp)
    : p(p), ibp(ibp)
{ }

/** Perform array reference. This implements the special abilities of starbase array properties:
    - index 0 counts all items
    - index 1..limit returns that item's count
    - other values yield empty (not an error!)
    \param func GPlanet function to access the property in raw form (e.g. getBaseEngineStore)
    \param limit Maximum index (e.g. NUM_ENGINES)
    \param arg User-specified argument
    \return resulting value */
IntValue*
IntBaseArrayProperty::performArrayReference(mp16_t (GPlanet::*func)(int) const, int32_t limit, int32_t arg, bool hull)
{
    if (arg == 0) {
        int32_t sum = 0;
        for (int i = 1; i <= limit; ++i) {
            mp16_t val = (p.*func)(i);
            if (!val.isKnown()) {
                return 0;
            }
            sum += val;
        }
        return makeIntValue(sum);
    } else if (hull) {
        if (arg < 0 || arg > getNumHulls())
            return 0;
        if (int n = getTruehullSlot(p.getOwner(), arg))
            return makeIntFromMaybe((p.*func)(n));
        else
            return makeIntValue(0);
    } else if (arg > 0 && arg <= limit) {
        return makeIntFromMaybe((p.*func)(arg));
    } else {
        return 0;
    }
}

IntValue*
IntBaseArrayProperty::get(IntArgBlock& args)
{
    int32_t arg;
    args.checkArgumentCount(1);
    if (!checkIntArg(arg, args.getNext()))
        return 0;

    switch (ibp) {
     case ibpEngineStorage:
        return performArrayReference(&GPlanet::getBaseEngineStore, NUM_ENGINES, arg, false);
     case ibpHullStorage:
        return performArrayReference(&GPlanet::getBaseHullStoreSlot, NUM_HULLS_PER_PLAYER, arg, true);
     case ibpBeamStorage:
        return performArrayReference(&GPlanet::getBaseBeamStore, NUM_BEAMS, arg, false);
     case ibpLauncherStorage:
        return performArrayReference(&GPlanet::getBaseLauncherStore, NUM_TORPS, arg, false);
     case ibpAmmoStorage:
        return performArrayReference(&GPlanet::getBaseAmmoStore, NUM_TORPS+1, arg, false);
     default:
        return 0;
    }
}

void
IntBaseArrayProperty::set(IntArgBlock& /*args*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

int32_t
IntBaseArrayProperty::getDimension(int32_t dim)
{
    if (dim == 0) {
        // 1 dimension
        return 1;
    } else {
        // Get that dimension
        // Note: must add "+1" because nonexistant 0th element is counted as well!
        // "dim a(10)" produces an array that has "dim(a) = 10", and indexes 0..9.
        switch (ibp) {
         case ibpEngineStorage:
            return NUM_ENGINES+1;
         case ibpHullStorage:
            return getNumHulls()+1;
         case ibpBeamStorage:
            return NUM_BEAMS+1;
         case ibpLauncherStorage:
            return NUM_TORPS+1;
         case ibpAmmoStorage:
            return NUM_TORPS+1+1;
         default:
            return 0;
        }
    }
}

IntContext*
IntBaseArrayProperty::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

IntBaseArrayProperty*
IntBaseArrayProperty::clone()
{
    return new IntBaseArrayProperty(p, ibp);
}

string_t
IntBaseArrayProperty::toString(bool /*readable*/)
{
    return "#<array>";
}

void
IntBaseArrayProperty::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}


/** Get starbase property.
    \param pl Starbase
    \param ibp Starbase property to query */
IntValue*
getBaseProperty(const GPlanet& pl, GUniverse& univ, IntBaseProperty ibp)
{
    if (!pl.isPlayable(GObject::Playable) || !pl.hasBase())
        return 0;

    switch (ibp) {
     case ibpBaseDamage:
        /* @q Damage:Int (Planet Property)
           Starbase damage. EMPTY if no base. */
        return makeIntFromMaybe(pl.getBaseDamage());
     case ibpBaseDefense:
        /* @q Defense.Base:Int (Planet Property)
           Starbase defense. EMPTY if no base. */
        return makeIntValue(pl.getStructures(str_BaseDefense));
     case ibpBaseDefenseMax:
        /* @q Defense.Base.Max:Int (Planet Property)
           Maximum starbase defense allowed. EMPTY if no base.
           @since PCC 1.1.16, PCC2 1.99.8 */
        return makeIntValue(getMaxStructures(pl, str_BaseDefense));
     case ibpBaseFighters:
        /* @q Fighters:Int (Planet Property)
           Number of fighters on starbase. EMPTY if no base. */
        return makeIntFromMaybe(pl.getCargo(el_Fighters));
     case ibpBaseFightersMax:
        /* @q Fighters.Max:Int (Planet Property)
           Maximum number of fighters allowed on starbase. EMPTY if no base.
           @since PCC 1.1.16, PCC2 1.99.8 */
        return makeIntValue(config.MaximumFightersOnBase(pl.getOwner()));
     case ibpBeamTech:
        /* @q Tech.Beam:Int (Planet Property)
           Beam tech level on starbase. EMPTY if no base. */
        return makeIntFromMaybe(pl.getBaseTechLevel(tl_Beam));
     case ibpBuildBeam:
        /* @q Build.Beam$:Int (Planet Property)
           Beam type for ship to build on starbase. EMPTY if no base. */
        return makeIntFromMaybe(mp16_t(pl.getBaseBuildOrder().beam_type));
     case ibpBuildBeamCount:
        /* @q Build.Beam.Count:Int (Planet Property)
           Number of beams for ship to build on starbase. EMPTY if no base. */
        return makeIntFromMaybe(mp16_t(pl.getBaseBuildOrder().beam_count));
     case ibpBuildEngine:
        /* @q Build.Engine$:Int (Planet Property)
           Number of engines for ship to build on starbase. EMPTY if no base. */
        return makeIntValue(mp16_t(pl.getBaseBuildOrder().engine));
     case ibpBuildFlag:
        /* @q Build.YesNo:Bool (Planet Property)
           Ship build flag. True if this base is building a ship. EMPTY if no base. */
        return makeBoolValue(mp16_t(pl.getBaseBuildHull() != 0));
     case ibpBuildHull:
        /* @q Build.Hull$:Int (Planet Property)
           Type of ship (hull Id) to build on starbase. EMPTY if no base, or no ship being built. */
        if (pl.getBaseBuildHull().isKnown() && pl.getBaseBuildHull() > 0)
            return makeIntValue(pl.getBaseBuildHull());
        else
            return 0;
     case ibpBuildHullName:
        /* @q Build:Str (Planet Property)
           Type of ship (hull name) to build on starbase. EMPTY if no base, or no ship being built.
           @see Name (Hull Property) */
        if (pl.getBaseBuildHull().isKnown()) {
            if (int n = pl.getBaseBuildHull()) {
                return makeStringValue(::getHull(n).getName());
            }
        }
        return 0;
     case ibpBuildHullShort:
        /* @q Build.Short:Str (Planet Property)
           Type of ship (short hull name) to build on starbase. EMPTY if no base, or no ship being built.
           @see Name.Short (Hull Property) */
        if (pl.getBaseBuildHull().isKnown()) {
            if (int n = pl.getBaseBuildHull()) {
                return makeStringValue(::getHull(n).getShortName());
            }
        }
        return 0;
     case ibpBuildQueuePos:
        /* @q Build.QPos:Int (Planet Property)
           Position of starbase in build queue. EMPTY if no base, or position not known. */
        return makeIntFromMaybe(pl.getBaseQueuePosition());
     case ibpBuildTorp:
        /* @q Build.Torp$:Int (Planet Property)
           Torpedo type for ship to build on starbase. EMPTY if no base. */
        return makeIntFromMaybe(mp16_t(pl.getBaseBuildOrder().torp_type));
     case ibpBuildTorpCount:
        /* @q Build.Torp.Count:Int (Planet Property)
           Number of torpedo tubes for ship to build on starbase. EMPTY if no base. */
        return makeIntFromMaybe(mp16_t(pl.getBaseBuildOrder().torp_count));
     case ibpEngineTech:
        /* @q Tech.Engine:Int (Planet Property)
           Engine tech level on starbase. EMPTY if no base. */
        return makeIntFromMaybe(pl.getBaseTechLevel(tl_Engine));
     case ibpHullTech:
        /* @q Tech.Hull:Int (Planet Property)
           Hull tech level on starbase. EMPTY if no base. */
        return makeIntFromMaybe(pl.getBaseTechLevel(tl_Hull));
     case ibpMission:
        /* @q Mission$:Int (Planet Property)
           Starbase mission number. EMPTY if no base.
           @assignable */
        return makeIntFromMaybe(pl.getBaseMission());
     case ibpMissionName:
        /* @q Mission:Str (Planet Property)
           Starbase mission. EMPTY if no base. */
        {
            mp16_t msn = pl.getBaseMission();
            if (msn.isKnown()) {
                return makeStringValue(getBaseMissionName(msn));
            } else {
                return 0;
            }
        }
     case ibpShipyardAction:
        /* @q Shipyard.Action:Str (Planet Property)
           Shipyard action on base. One of <tt>"Fix"</tt> or <tt>"Recycle"</tt>.
           EMPTY if no base, or no shipyard order set. */
        if (pl.getBaseShipyardOrder().isKnown()) {
            switch (pl.getBaseShipyardOrder()) {
             case sya_Fix:     return makeStringValue("Fix");
             case sya_Recycle: return makeStringValue("Recycle");
            }
        }
        return 0;
     case ibpShipyardId:
        /* @q Shipyard.Id:Int (Planet Property)
           Id of ship being worked on by starbase. EMPTY if no base. */
        return makeIntFromMaybe(pl.getBaseShipyardId());
     case ibpShipyardName:
        /* @q Shipyard.Name:Str (Planet Property)
           Name of ship being worked on by starbase. EMPTY if no base, or no shipyard order set. */
        if (pl.getBaseShipyardId().isKnown()) {
            if (int sid = pl.getBaseShipyardId()) {
                return makeStringValue(univ.getShip(sid).getName(GObject::PlainName));
            }
        }
        return 0;
     case ibpShipyardStr:
        /* @q Shipyard:Str (Planet Property)
           Shipyard order in human-readable form.
           A combination of {Shipyard.Action} and {Shipyard.Name}.
           EMPTY if no base, or no shipyard order set. */
        if (pl.getBaseShipyardOrder().isKnown() && pl.getBaseShipyardId().isKnown()) {
            switch (pl.getBaseShipyardOrder()) {
             case sya_Fix:     return makeStringValue(format("Fix %s", univ.getShip(pl.getBaseShipyardId()).getName(GObject::PlainName)));
             case sya_Recycle: return makeStringValue(format("Recycle %s", univ.getShip(pl.getBaseShipyardId()).getName(GObject::PlainName)));
            }
        }
        return 0;
     case ibpTorpedoTech:
        /* @q Tech.Torpedo:Int (Planet Property)
           Torpedo tech level on starbase. EMPTY if no base. */
        return makeIntFromMaybe(pl.getBaseTechLevel(tl_Torpedo));

     case ibpEngineStorage:
        /* @q Storage.Engines:Int() (Planet Property)
           Number of engines in starbase storage.
           Index can be 0 (=total number of engines) or an engine type (=number of engines of that type).
           EMPTY if no base. */
     case ibpHullStorage:
        /* @q Storage.Hulls:Int() (Planet Property)
           Number of engines in starbase storage.
           Index can be 0 (=total number of hulls) or a hull type (=number of engines of that type).
           EMPTY if no base. */
     case ibpBeamStorage:
        /* @q Storage.Beams:Int() (Planet Property)
           Number of beams in starbase storage.
           Index can be 0 (=total number of beams) or a beam type (=number of beams of that type).
           EMPTY if no base. */
     case ibpLauncherStorage:
        /* @q Storage.Launchers:Int() (Planet Property)
           Number of torpedo launchers in starbase storage.
           Index can be 0 (=total number of launchers) or a torpedo type (=number of launchers of that type).
           EMPTY if no base. */
     case ibpAmmoStorage:
        /* @q Storage.Ammo:Int() (Planet Property)
           Number of torpedoes or fighters in starbase storage.
           Index can be 0 (=total number of weapons), a torpedo type (=number of torpedoes of that type),
           or 11 (=number of fighters, see {Fighters (Planet Property)}.
           EMPTY if no base. */
        return new IntBaseArrayProperty(pl, ibp);
    }
    return 0;
}

void
setBaseProperty(GPlanet& pl, IntBaseProperty ibp, IntValue* value)
{
    if (!pl.hasBase() || !pl.isPlayable(GObject::Playable))
        throw IntError::notAssignable();

    int32_t iv;
    switch (ibp) {
     case ibpMission:
        // FIXME: we don't have a constant for '6'.
        if (checkIntArg(iv, value, 0, 6)) {
            pl.setBaseMission(mp16_t(iv));
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}

static void
setBaseOrder(GUniverse& univ, TShipyardAction order, GPlanet& pl, int sid)
{
    // Verify
    if (sid == 0) {
        // We want to cancel the order. So all we need is a base.
        mustHavePlayedBase(pl);
        order = sya_None;
    } else {
        // Verify order.
        if (GError e = checkBaseAction(order, univ, sid, pl.getId()))
            throw e;
    }

    // Do it
    pl.setBaseShipyardOrder(mp16_t(order), mp16_t(sid));
}

/** Backend for BuildXXX component build commands.
    \param pl   Planet to work on
    \param exc  Execution context (required to assign BUILD.REMAINDER)
    \param area Kind of component to build
    \param type Type of component to build (index, 1..NUM_HULLS_PER_PLAYER/NUM_ENGINES/...)
    \param amount Amount to build or scrap
    \param flag true to permit partial builds */
static void
doBaseBuildComponents(GPlanet& pl, IntExecutionContext& exc, TTechLevel area, int type, int amount, bool flag)
{
    // Do it (checks preconditions)
    GStarbaseBuildPartsAction action(pl);
    action.addUndoInformation(getDisplayedTurn().getPreviousUniverse().getPlanet(pl.getId()));
    int32_t did = action.addParts(area, type, amount, flag);

    // Now (try to) do it
    GMultiTransaction mt;
    GPlanetTransfer& tx           = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GStarbaseBuildTransaction& ta = mt.addNewSubtransaction(new GStarbaseBuildTransaction(action, tx));

    if (flag != 0) {
        // Try to back out if we have too little cash, then commit (this will throw if there is a different error).
        while (did > 0 && !ta.isValidCost() && action.addParts(area, type, -1, false))
            --did;
        mt.doCommit();

        // Tell user about remainder
        IntIntIntValue iv(amount - did);
        exc.setVariable("BUILD.REMAINDER", &iv);
    } else {
        // We are not permitted to do a partial build.
        if (did != amount) {
            mt.doCancel();
            throw GError(GError::eNoResource, GError::eNoResource);
        } else {
            mt.doCommit();
        }
    }
}

/** Backend for BuildXXX ammo build commands.
    \param pl   Planet to work on
    \param univ Universe to work in
    \param exc  Execution context (required to assign BUILD.REMAINDER)
    \param type Type of component to build
    \param amount Amount to build or scrap
    \param flag true to permit partial builds
    \param sid  Target ship Id */
static void
doBaseBuildAmmo(GPlanet& pl, GUniverse& univ, IntExecutionContext& exc, int type, int amount, bool flag, int32_t sid)
{
    // Do we have a base?
    mustHavePlayedBase(pl);

    // Do we have a valid ship Id?
    if (sid < 0 || sid > NUM_SHIPS)
        throw IntError::rangeError();
    if (sid > 0) {
        const GShip& sh = univ.getShip(sid);
        if (!sh.isPlayable(GObject::Playable) || sh.getPos() != pl.getPos())
            throw GError(GError::ePos, GError::ePos);
    }

    // Build objects.
    GMultiTransaction mt;
    GPlanetTransfer& payAmmo       = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GPlanetTransfer& payTech       = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    GCargoContainer& receiver      = (sid > 0
                                      ? static_cast<GCargoContainer&>(mt.addNewSubtransaction(new GShipTransfer(univ.getShip(sid))))
                                      : static_cast<GCargoContainer&>(mt.addNewSubtransaction(new GPlanetTransfer(pl))));
    GStarbaseBuildAmmoAction& ammo = mt.addNewSubtransaction(new GStarbaseBuildAmmoAction(payAmmo, receiver));
    GStarbaseTechUpgradeAction tech(pl);
    mt.addNewSubtransaction(new GStarbaseBuildTransaction(tech, payTech));
    // Phew.

    // If we're going to build something, so make sure we have the tech
    if (amount > 0) {
        int needed = ammo.getTechUpgradeNeeded(type);
        if (needed) {
            if (!tech.setTech(tl_Torpedo, needed))
                throw GError(GError::ePerm, GError::ePerm);
            if (GError e = mt.getError())
                throw e;
        }
    }

    // Build
    int32_t n = ammo.limitAmount(type, amount);
    if (flag) {
        ammo.setAmount(type, ammo.getAmount(type) + n);
        mt.doCommit();

        IntIntIntValue iv(amount - n);
        exc.setVariable("BUILD.REMAINDER", &iv);
    } else {
        if (n != amount)
            throw GError(GError::eNoResource, GError::eNoResource);
        ammo.setAmount(type, ammo.getAmount(type) + n);
        mt.doCommit();
    }
}

/* @q BuildFighters amount:Int, Optional flagAndShipId:Any (Planet Command)
   Build fighters.

   Attempts to build %amount fighters. The amount can be negative to
   scrap fighters. The %flagAndShipId can be "N" to permit partial
   builds. If not all of the requested amount can be built, the
   command will report the amount not built in the variable
   %Build.Remainder instead of failing.

   %flagAndShipId can also contain a ship Id, to place the
   newly-built fighters on that ship.

   @since PCC2 1.99.9, PCC 1.1.5 */
void
IFBaseBuildFighters(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(1, 2);

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, 10000))
        return;

    int32_t flag = 0;
    int32_t sid = 0;
    checkFlagArg(flag, &sid, args.getNext(), "N");

    doBaseBuildAmmo(static_cast<GPlanet&>(obj), univ, exc, GStarbaseBuildAmmoAction::max_index, amount, flag, sid);
}

/* @q BuildEngines type:Int, amount:Int, Optional flag:Str (Planet Command)
   Build engines.

   Attempts to build %amount engines of the given %type. The amount
   can be negative to scrap engines. The tech levels is automatically
   raised as necessary. The %flag can be "N" to permit partial
   builds. If not all of the requested amount can be built, the
   command will report the amount not built in the variable
   %Build.Remainder instead of failing.

   @since PCC2 1.99.9, PCC 1.1.16 */
void
IFBaseBuildEngines(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, NUM_ENGINES))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, +10000))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    doBaseBuildComponents(static_cast<GPlanet&>(obj), exc, tl_Engine, type, amount, flag);
}

/* @q BuildHulls type:Int, amount:Int, Optional flag:Str (Planet Command)
   Build starship hulls.

   Attempts to build %amount hulls of the given %type. The amount
   can be negative to scrap hulls. The tech levels is automatically
   raised as necessary. The %flag can be "N" to permit partial
   builds. If not all of the requested amount can be built, the
   command will report the amount not built in the variable
   %Build.Remainder instead of failing.

   The %type is a hull Id. You can not build all hulls; the command will
   fail if you try to build one you cannot build.

   @since PCC2 1.99.9, PCC 1.1.16 */
void
IFBaseBuildHulls(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, getNumHulls()))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, +10000))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    // Can we build this hull?
    int32_t slot = getTruehullSlot(obj.getOwner(), type);
    if (!slot) {
        if (amount != 0)
            throw GError(GError::ePerm, GError::ePerm);
    } else {
        doBaseBuildComponents(static_cast<GPlanet&>(obj), exc, tl_Hull, slot, amount, flag);
    }
}

/* @q BuildLaunchers type:Int, amount:Int, Optional flag:Str (Planet Command)
   Build torpedo launchers.

   Attempts to build %amount torpedo launchers of the given %type.
   The amount can be negative to scrap launchers. The tech levels is
   automatically raised as necessary. The %flag can be "N" to permit
   partial builds. If not all of the requested amount can be built,
   the command will report the amount not built in the variable
   %Build.Remainder instead of failing.

   @since PCC2 1.99.9, PCC 1.1.16 */
void
IFBaseBuildLaunchers(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, NUM_TORPS))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, +10000))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    doBaseBuildComponents(static_cast<GPlanet&>(obj), exc, tl_Torpedo, type, amount, flag);
}

/* @q BuildBeams type:Int, amount:Int, Optional flag:Str (Planet Command)
   Build beam weapons.

   Attempts to build %amount beams of the given %type. The amount
   can be negative to scrap beams. The tech levels is automatically
   raised as necessary. The %flag can be "N" to permit partial
   builds. If not all of the requested amount can be built, the
   command will report the amount not built in the variable
   %Build.Remainder instead of failing.

   @since PCC2 1.99.9, PCC 1.1.16 */
void
IFBaseBuildBeams(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, NUM_BEAMS))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, +10000))
        return;

    int32_t flag = 0;
    checkFlagArg(flag, 0, args.getNext(), "N");

    doBaseBuildComponents(static_cast<GPlanet&>(obj), exc, tl_Beam, type, amount, flag);
}

/* @q BuildTorps type:Int, amount:Int, Optional flagAndShipId:Any (Planet Command)
   Build torpedoes.

   Attempts to build %amount torpedoes of the given type. The amount
   can be negative to scrap torpedoes. The %flagAndShipId can be "N"
   to permit partial builds. If not all of the requested amount can
   be built, the command will report the amount not built in the
   variable %Build.Remainder instead of failing.

   %flagAndShipId can also contain a ship Id, to place the
   newly-built torpedoes on that ship.

   @since PCC2 1.99.9, PCC 1.1.5 */
void
IFBaseBuildTorps(GObject& obj, GUniverse& univ, IntExecutionContext& exc, IntArgBlock& args)
{
    args.checkArgumentCount(2, 3);

    int32_t type;
    if (!checkIntArg(type, args.getNext(), 1, NUM_TORPS))
        return;

    int32_t amount;
    if (!checkIntArg(amount, args.getNext(), -10000, 10000))
        return;

    int32_t flag = 0;
    int32_t sid = 0;
    checkFlagArg(flag, &sid, args.getNext(), "N");

    doBaseBuildAmmo(static_cast<GPlanet&>(obj), univ, exc, type, amount, flag, sid);
}

/** Parse build ship command.
    \param args [in] arguments
    \param o [out] build order
    \retval true correctly parsed
    \retval false null argument encountered
    \throws GError or IntError on error */
bool
parseBuildShipCommand(IntArgBlock& args, TBuildOrder& o)
{
    args.checkArgumentCount(1, 6);

    // Mandatory arg
    int32_t hull;
    if (!checkIntArg(hull, args.getNext(), 0, getNumHulls())) {
        return false;
    }

    // Optional args
    int32_t engine = 0, beam = 0, beamcount = -1, torp = 0, torpcount = -1;
    checkIntArg(engine, args.getNext(), 1, NUM_ENGINES);
    checkIntArg(beam, args.getNext(), 0, NUM_BEAMS);
    checkIntArg(beamcount, args.getNext());
    checkIntArg(torp, args.getNext(), 0, NUM_BEAMS);
    checkIntArg(torpcount, args.getNext());

    // Check mandatory arg
    o.hull = hull;
    if (hull == 0) {
        // Canceling a build
        return true;
    }

    // This is a ship build order. Validate remaining args
    const GHull& h = getHull(hull);

    // Engine
    if (engine <= 0)
        throw IntError::rangeError();
    o.engine = engine;

    // Beams
    if (beamcount == -1)
        beamcount = h.getMaxBeams();
    if (beam == 0 || beamcount == 0) {
        o.beam_type = o.beam_count = 0;
    } else if (beamcount < 0 || beamcount > h.getMaxBeams()) {
        throw IntError::rangeError();
    } else {
        o.beam_type  = beam;
        o.beam_count = beamcount;
    }

    // Torps
    if (torpcount == -1)
        torpcount = h.getMaxLaunchers();
    if (torp == 0 || torpcount == 0) {
        o.torp_type = o.torp_count = 0;
    } else if (torpcount < 0 || torpcount > h.getMaxLaunchers()) {
        throw IntError::rangeError();
    } else {
        o.torp_type  = torp;
        o.torp_count = torpcount;
    }
    return true;
}

/* @q BuildShip hull:Int, Optional engine:Int, beamtype:Int, beamcount:Int, torptype:Int, torpcount:Int (Planet Command)
   Submit a starship build order.
   If %hull is zero, cancels a pending order.
   Otherwise, builds a ship.
   In this case, %engine must be specified, and the others should better be specified as well to
   avoid building a ship without weapons.

   Required tech levels and parts are bought automatically.

   @since PCC2 1.99.16, PCC 1.0.6 */
void
IFBaseBuildShip(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    // Parse args
    TBuildOrder o;
    if (!parseBuildShipCommand(args, o))
        return;

    // Get planet
    GPlanet& pl = static_cast<GPlanet&>(obj);
    mustHavePlayedBase(pl);

    // Check for cancellation
    if (o.hull == 0) {
        pl.setBaseBuildOrder(o);
        return;
    }

    // Make a transaction and fire it
    GStarbaseBuildShipAction action(pl);
    action.setUsePartsFromStorage(true);
    action.setUseTechUpgrade(true);
    action.setBuildOrder(o);

    GMultiTransaction mt;
    GPlanetTransfer& tx = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    mt.addNewSubtransaction(new GStarbaseBuildTransaction(action, tx));
    mt.doCommit();
    pl.setBaseBuildOrder(action.getBuildOrder()); // FIXME: why is this separate?
}

/* @q FixShip sid:Int (Planet Command)
   Fix (repair) a ship. The %sid is a ship Id, or 0 to cancel a pending shipyard order.
   @since PCC2 1.99.9, PCC 1.0.5 */
void
IFBaseFixShip(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t sid;
    args.checkArgumentCount(1);
    if (checkIntArg(sid, args.getNext(), 0, NUM_SHIPS))
        setBaseOrder(univ, sya_Fix, static_cast<GPlanet&>(obj), sid);
}

/* @q RecycleShip sid:Int (Planet Command)
   Recycle a ship. The %sid is a ship Id, or 0 to cancel a pending shipyard order.
   @since PCC2 1.99.9, PCC 1.0.5 */
void
IFBaseRecycleShip(GObject& obj, GUniverse& univ, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    int32_t sid;
    args.checkArgumentCount(1);
    if (checkIntArg(sid, args.getNext(), 0, NUM_SHIPS))
        setBaseOrder(univ, sya_Recycle, static_cast<GPlanet&>(obj), sid);
}

/* @q SetMission number:Int (Planet Command)
   Set starbase mission.
   @since PCC2 1.99.9, PCC 1.0.5 */
void
IFBaseSetMission(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    setBaseProperty(static_cast<GPlanet&>(obj), ibpMission, args.getNext());
}

/* @q SetTech area:Int, level:Int (Planet Command)
   Set starbase tech level. %area is 1 for engines, 2 for hulls, 3
   for beams, 4 for torpedoes. %level is the new tech level.

   Note that when you build a ship, PCC automatically upgrades tech.
   You can raise tech levels, and lower them again when you have not
   yet used them.

   @since PCC2 1.99.9, PCC 1.1 */
void
IFBaseSetTech(GObject& obj, GUniverse& /*univ*/, IntExecutionContext& /*exc*/, IntArgBlock& args)
{
    args.checkArgumentCount(2);

    // Fetch arguments
    int32_t area, tech;
    if (!checkIntArg(area, args.getNext(), 1, 4))
        return;
    if (!checkIntArg(tech, args.getNext(), 1, 10))
        return;
    --area;

    // Get planet
    GPlanet& pl = static_cast<GPlanet&>(obj);

    // Create tech upgrade action (checks preconditions)
    GStarbaseTechUpgradeAction action(pl);
    GUndoInformation undo;
    undo.computeInformation(getDisplayedTurn(), pl.getId());
    action.setUndoInformation(undo);

    // Do the rules permit this?
    if (!action.setTech(area, tech))
        throw GError(GError::ePerm, GError::ePerm);

    // Now (try to) do it
    GMultiTransaction mt;
    GPlanetTransfer& tx = mt.addNewSubtransaction(new GPlanetTransfer(pl));
    mt.addNewSubtransaction(new GStarbaseBuildTransaction(action, tx));
    mt.doCommit();
}
